//  ---------------------------------------------------------------------------
//  This file is part of reSID, a MOS6581 SID emulator engine.
//  Copyright (C) 2004  Dag Lem <resid@nimrod.no>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ---------------------------------------------------------------------------

#ifndef VICE__ENVELOPE_H__
#define VICE__ENVELOPE_H__

#include "siddefs-fp.h"

// ----------------------------------------------------------------------------
// A 15 bit counter is used to implement the envelope rates, in effect
// dividing the clock to the envelope counter by the currently selected rate
// period.
// In addition, another counter is used to implement the exponential envelope
// decay, in effect further dividing the clock to the envelope counter.
// The period of this counter is set to 1, 2, 4, 8, 16, 30 at the envelope
// counter values 255, 93, 54, 26, 14, 6, respectively.
// ----------------------------------------------------------------------------
typedef enum ADSRState { ATTACK, DECAY_SUSTAIN, RELEASE } ADSRState;
typedef struct EnvelopeGeneratorFP
{
  int rate_counter;
  int rate_period;
  reg8 exponential_counter;
  reg8 exponential_counter_period;
  reg8 envelope_counter;
  reg8 muted, hold_zero;

  reg8 attack;
  reg8 decay;
  reg8 sustain;
  reg8 release;

  reg8 gate;

  ADSRState state;

  // Lookup table to convert from attack, decay, or release value to rate
  // counter period.
  //static reg16 rate_counter_period[];

  // The 16 selectable sustain levels.
  //static reg8 sustain_level[];
} EnvFP, EnvelopeGeneratorFP;

#ifdef __cplusplus
extern "C" {
#endif

void update_rate_period(EnvFP *e, reg16 period);
void EnvWriteCONTROL_REG(EnvFP *e, reg8 gate_next);
void EnvClock(EnvFP *e);

// ----------------------------------------------------------------------------
// SID reset.
// ----------------------------------------------------------------------------
static inline void EnvReset(EnvFP *e)
{
  e->rate_counter = 0;
  //e->rate_period = rate_counter_period[0];
  e->rate_period = 9;
  e->attack = e->decay = e->sustain = e->release = 0;
  e->gate = 0;
  e->state = RELEASE;
  e->envelope_counter = 0;
  e->hold_zero = 0;
  e->muted = 0;

  e->exponential_counter = 0;
  e->exponential_counter_period = 1;
}

static inline void EnvWriteATTACK_DECAY(EnvFP *e, reg8 ad)
{
  e->attack = ((ad >> 4) & 0x0f);
  e->decay = (ad & 0x0f);
  if (e->state == ATTACK) { ad >>= 4; }
  if (e->state != RELEASE) {
    update_rate_period(e, ad & 0x0f);
  }
}

static inline void EnvWriteSUSTAIN_RELEASE(EnvFP *e, reg8 sr)
{
  e->sustain = ((sr >> 4) & 0x0f);
  e->release = (sr & 0x0f);
  if (e->state == RELEASE) {
    update_rate_period(e, e->release);
  }
}

static inline void EnvMute(EnvFP *e, reg8 enable)
{
  e->muted = !!enable;
}

// ----------------------------------------------------------------------------
// Read the envelope generator output.
// ----------------------------------------------------------------------------
static inline reg8 EnvRead(EnvFP *e)
{
  return e->muted ? 0 : e->envelope_counter;
}

#ifdef __cplusplus
};
#endif

#endif // not VICE__ENVELOPE_H__
