//  ---------------------------------------------------------------------------
//  This file is part of reSID, a MOS6581 SID emulator engine.
//  Copyright (C) 2004  Dag Lem <resid@nimrod.no>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ---------------------------------------------------------------------------
//  Filter distortion code written by Antti S. Lankila 2007 - 2009.

#include <stdint.h>
#include <math.h>
#include "filter_c.h"
extern float expf(float val);
 
// ----------------------------------------------------------------------------
// Constructor.
// ----------------------------------------------------------------------------
void filter_init(FilterFP *s)
{
  s->enabled = 1;
  s->model = MOS6581FP;
  s->distortion_nonlinearity = 3.3e6f;
  filter_set_nonlinearity(s, 0.96f);
  s->clock_frequency = 1000000.0f;
  s->distortion_CT = 1.f / (sidcaps_6581 * s->clock_frequency);
  filter_set_type3_properties(s, 129501.5675945764f, 284015710.29875594f, 1.006508973f, 18741.32407361f);
  s->type4_k = 20;
  s->type4_b = 6.55f;
  s->type4_w0_cache = 2 * (float)M_PI / s->clock_frequency;
  filter_reset(s);
}


// ----------------------------------------------------------------------------
// Enable filter.
// ----------------------------------------------------------------------------
void filter_enable(FilterFP *s, unsigned int enable)
{
  s->enabled = enable;
  if (! enable)
    s->filt = 0; // XXX should also restore this...
}

// ----------------------------------------------------------------------------
// Set chip model.
// ----------------------------------------------------------------------------
void filter_set_chip_model(FilterFP *s, chip_model model)
{
    s->model = model;
}

void set_distortion_properties(FilterFP *s, float a, float nl)
{
    s->distortion_nonlinearity = nl;
}

// Set filter cutoff frequency.
void set_w0(FilterFP *s)
{
  const float nl = s->nonlinearity;
  const float dir = 2.0f * nl;
  float value = 0.0f, weight = 1.0f;
  int i, bit = 1;
  for (i = 0; i < 11; i ++) {
      if (s->fc & bit)
          value += weight;
      bit <<= 1;
      weight *= dir;
  }

  s->type3_fc_kink = value / (weight / nl / nl ) * 2048.0f;
}

void filter_set_nonlinearity(FilterFP *s, float nl)
{
    s->nonlinearity = nl;
    set_w0(s);
}

/* dist_CT eliminates 1/x at hot spot */
void filter_set_clock_frequency(FilterFP *s, float clock) {
    s->clock_frequency = clock;
    s->distortion_CT = 1.f / (sidcaps_6581 * clock);
    s->type4_w0_cache = 2.0f * (float)M_PI / clock;
    //set_w0(s);
}

void filter_set_type4_properties(FilterFP *s, float k, float b)
{
    s->type4_k = 2*k;
    s->type4_b = 2*b;
}

void filter_set_type3_properties(FilterFP *s, float br, float o, float st, float mfr)
{
    st = st - 1.0f;
    s->type3_baseresistance = br;
    s->type3_offset = o;
    //type3_steepness = -logf(s) / 512.f; /* s^x to e^(x*ln(s)), 1/e^x == e^-x. */
    s->type3_steepness = -(st - 0.5f*st*st + st*st*st/3.0f) / 512.f; /* s^x to e^(x*ln(s)), 1/e^x == e^-x. */
    s->type3_minimumfetresistance = mfr;
    s->type3_fc_kink_exp = s->type3_offset * expf(s->type3_fc_kink * s->type3_steepness * 512.f);
}

float filter_type3_w0(FilterFP *s, const float dist)
{
    float fetresistance = s->type3_fc_kink_exp;
    if (dist > 0) {
        fetresistance *= fastexp(dist * s->type3_steepness);
    }
    const float dynamic_resistance = s->type3_minimumfetresistance + fetresistance;

    /* 2 parallel resistors */
    const float _1_div_resistance = (s->type3_baseresistance + dynamic_resistance) / (s->type3_baseresistance * dynamic_resistance);
    /* 1.f / (clock * caps * resistance) */
    return s->distortion_CT * _1_div_resistance;
}

float filter_type4_w0(FilterFP *s)
{
    const float freq = s->type4_k * s->fc + s->type4_b;
    return freq * s->type4_w0_cache;
}

// ----------------------------------------------------------------------------
// SID reset.
// ----------------------------------------------------------------------------
void filter_reset(FilterFP *s)
{
  s->fc = 0;
  s->res = s->filt = s->voice3off = s->hp_bp_lp = 0; 
  s->vol = 0;
  // State of filter.
  s->volf = s->eVhp = s->Vhp = s->Vbp = s->Vlp = 0;
  s->type3_fc_kink = s->type3_fc_kink_exp = 0;
  set_w0(s);
  s->_1_div_Q = 2.0f;
}

// ----------------------------------------------------------------------------
// Register functions.
// ----------------------------------------------------------------------------
void filter_write_FC_LO(FilterFP *s, reg8 fc_lo)
{
  s->fc = (s->fc & 0x7f8) | (fc_lo & 0x007);
  set_w0(s);
}

void filter_write_FC_HI(FilterFP *s, reg8 fc_hi)
{
  s->fc = ((fc_hi << 3) & 0x7f8) | (s->fc & 0x007);
  set_w0(s);
}

void filter_write_RES_FILT(FilterFP *s, reg8 res_filt)
{
  s->res = (res_filt >> 4) & 0x0f;
  s->_1_div_Q = 1.f / (0.5f + s->res / 18.f);
  s->filt = res_filt & 0x0f;
}

void filter_write_MODE_VOL(FilterFP *s, reg8 mode_vol)
{
  s->voice3off = mode_vol & 0x80;
  s->hp_bp_lp = (mode_vol >> 4) & 0x07;
  s->vol = mode_vol & 0x0f;
}

reg8 filter_read_RES_FILT(FilterFP *s)
{
  return ((s->res << 4) | s->filt);
}

reg8 filter_read_MODE_VOL(FilterFP *s)
{
    return ((s->voice3off ? 0x80 : 0) | (s->hp_bp_lp << 4) | s->vol);
}

// ----------------------------------------------------------------------------
// SID clocking - 1 cycle.
// ----------------------------------------------------------------------------
float filter_clock(FilterFP *s, float voice1, float voice2, float voice3, float ext_in)
{
    float dVhp, Vi = 0.f, Vf = 0.f;
    reg8 filt = s->filt;

    // Route voices into or around filter.
    if (filt & 1) Vi += voice1;
    else          Vf += voice1;
    if (filt & 2) Vi += voice2;
    else          Vf += voice2;
    //((filt & 1) ? Vi : Vf) += voice1;
    //((filt & 2) ? Vi : Vf) += voice2;
    // NB! Voice 3 is not silenced by voice3off if it is routed through
    // the filter.
    if (filt & 4) {
        Vi += voice3;
    } else if (! s->voice3off) {
        Vf += voice3;
    }
    //((filt & 8) ? Vi : Vf) += ext_in;
    if (filt & 8) Vi += ext_in;
    else          Vf += ext_in;

    if (s->hp_bp_lp & 1) {
        Vf += s->Vlp;
    }
    if (s->hp_bp_lp & 2) {
        Vf += s->Vbp;
    }
    if (s->hp_bp_lp & 4) {
        Vf += s->Vhp;
    }
    
    if (s->model == MOS6581FP) {
        s->Vlp -= s->Vbp * filter_type3_w0(s, s->Vbp);
        s->Vbp -= s->Vhp * filter_type3_w0(s, s->Vhp);
        s->Vhp = (s->Vbp * s->_1_div_Q - s->Vlp - Vi);
        Vf *= s->volf;
        //Vf = waveshaper1(Vf);
        if (Vf > s->distortion_nonlinearity) {
            Vf -= (Vf - s->distortion_nonlinearity) * 0.5f;
        }
    } else {
        /* On the 8580, BP appears mixed in phase with the rest. */
        s->Vlp += s->Vbp * filter_type4_w0(s);
        s->Vbp += s->Vhp * filter_type4_w0(s);
        s->Vhp = -s->Vbp * s->_1_div_Q - s->Vlp - Vi;
        Vf *= s->volf;
    }

    dVhp = 9.999983e-05f * (Vf - s->eVhp);
    s->eVhp += dVhp;
    return Vf - s->eVhp;
}

