//  ---------------------------------------------------------------------------
//  This file is part of reSID, a MOS6581 SID emulator engine.
//  Copyright (C) 2004  Dag Lem <resid@nimrod.no>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ---------------------------------------------------------------------------

#include <stdint.h>
#include <stdlib.h>
#include <math.h>
#include "sid.h"

static float SIDFP_kinked_dac(const int x, const float nonlinearity, const int max)
{
    float value = 0.f;

    int bit = 1;
    float weight = 1.f;
    const float dir = 2.0f * nonlinearity;
    for (int i = 0; i < max; i ++) {
        if (x & bit)
            value += weight;
        bit <<= 1;
        weight *= dir;
    }

    return value / (weight / (nonlinearity * nonlinearity)) * (1 << max);
}

/* tables used by voice/envgen */
static float env_dac[256];

/* nonlinear DAC support, set 1 for 8580 / no effect, about 0.96 otherwise */
void sid_set_voice_nonlinearity(SIDFP *sid, float nl)
{
  int i;
  //wave_set_nonlinearity(nl);
  //wave_rebuild_wftable(model);
  filter_set_nonlinearity(&sid->filter,nl);
  for (i = 0; i < 256; i ++)
    env_dac[i] = SIDFP_kinked_dac(i, nl, 8);
}

// ----------------------------------------------------------------------------
// Constructor.
// ----------------------------------------------------------------------------
void sid_init(SIDFP *sid)
{
  sid_set_sampling_parameters(sid, 985248, SAMPLE_INTERPOLATE, 44100);
  wave_set_nonlinearity(1.0f);
  sid_set_chip_model(sid,MOS6581FP);
  sid_reset(sid);

  sid_input(sid,0);
}


// ----------------------------------------------------------------------------
// Destructor.
// ----------------------------------------------------------------------------
void sid_free(SIDFP *sid)
{
}


// ----------------------------------------------------------------------------
// Set chip model.
// ----------------------------------------------------------------------------
void sid_set_chip_model(SIDFP *sid, chip_model model)
{
  int i;

  sid->model = model;

  for (i = 0; i < 3; i++) {
    wave_set_chip_model(&sid->wave[i], model);
  }
  wave_rebuild_wftable(model);

  if (sid->model == MOS6581FP) {
    /* there is some level from each voice even if the env is down and osc
     * is stopped. You can hear this by routing a voice into filter (filter
     * should be kept disabled for this) as the master level changes. This
     * tunable affects the volume of digis. */
    sid->voice_DC = (float)(0x800 * 0xff);
    /* In 8580 the waveforms seem well centered, but on the 6581 there is some
     * offset change as envelope grows, indicating that the waveforms are not
     * perfectly centered. The likely cause for this is the follows:
     *
     * The waveform DAC generates a voltage between 5 and 12 V corresponding
     * to oscillator state 0 .. 4095.
     *
     * The envelope DAC generates a voltage between waveform gen output and
     * the 5V level.
     *
     * The outputs are amplified against the 12V voltage and sent to the
     * mixer.
     *
     * The SID virtual ground is around 6.5 V. */
    sid_set_voice_nonlinearity(sid, 0.96f);
  }
  else {
    sid->voice_DC = 0.f;
    sid_set_voice_nonlinearity(sid, 1.0f);
  }
  filter_set_chip_model(&sid->filter, model);
}

// ----------------------------------------------------------------------------
// SID reset.
// ----------------------------------------------------------------------------
void sid_reset(SIDFP *sid)
{
  int i;
  for (i = 0; i < 3; i++) {
    EnvReset(&sid->envelope[i]);
    wave_reset(&sid->wave[i]);
  }

  filter_reset(&sid->filter);

  sid->bus_value = 0;
  sid->bus_value_ttl = 0;
  sid->Vo = 0.0f;
}


// ----------------------------------------------------------------------------
// Write 16-bit sample to audio input.
// NB! The caller is responsible for keeping the value within 16 bits.
// Note that to mix in an external audio signal, the signal should be
// resampled to 1MHz first to avoid sampling noise.
// ----------------------------------------------------------------------------
void sid_input(SIDFP *sid, int sample)
{
  // Voice outputs are 20 bits. Scale up to match three voices in order
  // to facilitate simulation of the MOS8580 "digi boost" hardware hack.
  sid->ext_in = (float)((sample << 4) * 3);
}

float sid_output(SIDFP *sid)
{
  /* Oscillators go from 0 to 4095, envelope from 0 to 255, volume goes to 1
   * and there are 3 voices. With strong resonance, it's still possible to
   * exceed the max (Drum Fool, for instance). So extra 50 % is allocated.
   *
   * Output range is -1 to 1. */
  float v = sid->Vo / (2047.f * 255.f * 3.f * 2.0f);
  if (v >  1.0f) v =  1.0f;
  if (v < -1.0f) v = -1.0f;
  return v;
}

// ----------------------------------------------------------------------------
// Read registers.
//
// Reading a write only register returns the last byte written to any SID
// register. The individual bits in this value start to fade down towards
// zero after a few cycles. All bits reach zero within approximately
// $2000 - $4000 cycles.
// It has been claimed that this fading happens in an orderly fashion, however
// sampling of write only registers reveals that this is not the case.
// NB! This is not correctly modeled.
// The actual use of write only registers has largely been made in the belief
// that all SID registers are readable. To support this belief the read
// would have to be done immediately after a write to the same register
// (remember that an intermediate write to another register would yield that
// value instead). With this in mind we return the last value written to
// any SID register for $2000 cycles without modeling the bit fading.
// ----------------------------------------------------------------------------
reg8 sid_read(SIDFP *sid, reg8 offset)
{
  if (offset == 0x1b) {
    return sid->model == MOS6581FP
        ? wave_readOSC6581(&sid->wave[2],&sid->wave[0])
        : wave_readOSC8580(&sid->wave[2],&sid->wave[0]);
  } else {
    return sid->bus_value;
  }
}

// ----------------------------------------------------------------------------
// Write registers.
// ----------------------------------------------------------------------------
void sid_write(SIDFP *sid, reg8 offset, reg8 value)
{
  sid->bus_value = value;
  sid->bus_value_ttl = 34000;

  switch (offset) {
  case 0x00:
    wave_writeFREQ_LO(&sid->wave[0],value);
    break;
  case 0x01:
    wave_writeFREQ_HI(&sid->wave[0],value);
    break;
  case 0x02:
    wave_writePW_LO(&sid->wave[0],value);
    break;
  case 0x03:
    wave_writePW_HI(&sid->wave[0],value);
    break;
  case 0x04:
    wave_writeCONTROL_REG(&sid->wave[0], &sid->wave[1], value);
    EnvWriteCONTROL_REG(&sid->envelope[0],value&1);
    break;
  case 0x05:
    EnvWriteATTACK_DECAY(&sid->envelope[0], value);
    break;
  case 0x06:
    EnvWriteSUSTAIN_RELEASE(&sid->envelope[0], value);
    break;
  case 0x07:
    wave_writeFREQ_LO(&sid->wave[1],value);
    break;
  case 0x08:
    wave_writeFREQ_HI(&sid->wave[1],value);
    break;
  case 0x09:
    wave_writePW_LO(&sid->wave[1],value);
    break;
  case 0x0a:
    wave_writePW_HI(&sid->wave[1],value);
    break;
  case 0x0b:
    wave_writeCONTROL_REG(&sid->wave[1], &sid->wave[2], value);
    EnvWriteCONTROL_REG(&sid->envelope[1],value&1);
    break;
  case 0x0c:
    EnvWriteATTACK_DECAY(&sid->envelope[1], value);
    break;
  case 0x0d:
    EnvWriteSUSTAIN_RELEASE(&sid->envelope[1], value);
    break;
  case 0x0e:
    wave_writeFREQ_LO(&sid->wave[2],value);
    break;
  case 0x0f:
    wave_writeFREQ_HI(&sid->wave[2],value);
    break;
  case 0x10:
    wave_writePW_LO(&sid->wave[2],value);
    break;
  case 0x11:
    wave_writePW_HI(&sid->wave[2],value);
    break;
  case 0x12:
    wave_writeCONTROL_REG(&sid->wave[2], &sid->wave[0], value);
    EnvWriteCONTROL_REG(&sid->envelope[2],value&1);
    break;
  case 0x13:
    EnvWriteATTACK_DECAY(&sid->envelope[2], value);
    break;
  case 0x14:
    EnvWriteSUSTAIN_RELEASE(&sid->envelope[2], value);
    break;
  case 0x15:
    filter_write_FC_LO(&sid->filter, value);
    break;
  case 0x16:
    filter_write_FC_HI(&sid->filter, value);
    break;
  case 0x17:
    filter_write_RES_FILT(&sid->filter, value);
    break;
  case 0x18:
    filter_write_MODE_VOL(&sid->filter, value);
    break;
  default:
    break;
  }
}


// ----------------------------------------------------------------------------
// Constructor.
// ----------------------------------------------------------------------------
/*SIDFP::State::State()
{
  int i;

  for (i = 0; i < 0x20; i++) {
    sid_register[i] = 0;
  }

  bus_value = 0;
  bus_value_ttl = 0;
}*/


// ----------------------------------------------------------------------------
// Read state.
// ----------------------------------------------------------------------------
void sid_read_state(SIDFP *sid, SIDFPState *state)
{
  int i, j;

  for (i = 0, j = 0; i < 3; i++, j += 7) {
    state->sid_register[j + 0] = sid->wave[i].freq & 0xff;
    state->sid_register[j + 1] = sid->wave[i].freq >> 8;
    state->sid_register[j + 2] = sid->wave[i].pw & 0xff;
    state->sid_register[j + 3] = sid->wave[i].pw >> 8;
    state->sid_register[j + 4] =
        (sid->wave[i].waveform << 4)
      | (sid->wave[i].test ? 0x08 : 0)
      | (sid->wave[i].ring ? 0x04 : 0)
      | (sid->wave[i].sync ? 0x02 : 0)
      | (sid->envelope[i].gate & 1);
    state->sid_register[j + 5] = ((sid->envelope[i].attack << 4) | sid->envelope[i].decay);
    state->sid_register[j + 6] = ((sid->envelope[i].sustain << 4) | sid->envelope[i].release);
  }

  state->sid_register[j++] = sid->filter.fc & 0x007;
  state->sid_register[j++] = sid->filter.fc >> 3;
  state->sid_register[j++] = filter_read_RES_FILT(&sid->filter);
  state->sid_register[j++] = filter_read_MODE_VOL(&sid->filter);

  // These registers are superfluous, but included for completeness.
  state->sid_register[j++] = 0xff;
  state->sid_register[j++] = 0xff;
  state->sid_register[j++] = sid_read(sid,0x1b);
  state->sid_register[j++] = sid->envelope[2].envelope_counter;
  state->sid_register[j++] = 0;
  state->sid_register[j++] = 0;

  state->bus_value = sid->bus_value;
  state->bus_value_ttl = sid->bus_value_ttl;
}


// ----------------------------------------------------------------------------
// Write state.
// ----------------------------------------------------------------------------
void sid_write_state(SIDFP *sid, const SIDFPState *state)
{
  sid_reset(sid);

  int i;
  for (i = 0; i <= 0x18; i++) {
    sid_write(sid, i, state->sid_register[i]);
  }

  sid->bus_value = state->bus_value;
  sid->bus_value_ttl = state->bus_value_ttl;
}


// ----------------------------------------------------------------------------
// SID voice muting.
// ----------------------------------------------------------------------------
void sid_mute(SIDFP *sid, reg8 channel, reg8 enable)
{
  // Only have 3 voices!
  if (channel >= 3)
    return;
  EnvMute (&sid->envelope[channel], enable);
}


// ----------------------------------------------------------------------------
// Enable filter.
// ----------------------------------------------------------------------------
void sid_enable_filter(SIDFP *sid, reg8 enable)
{
  filter_enable (&sid->filter, enable);
}


// ----------------------------------------------------------------------------
// Setting of SID sampling parameters.
//
// Use a clock freqency of 985248Hz for PAL C64, 1022730Hz for NTSC C64.
// The default end of passband frequency is pass_freq = 0.9*sample_freq/2
// for sample frequencies up to ~ 44.1kHz, and 20kHz for higher sample
// frequencies.
//
// For resampling, the ratio between the clock frequency and the sample
// frequency is limited as follows:
//   125*clock_freq/sample_freq < 16384
// E.g. provided a clock frequency of ~ 1MHz, the sample frequency can not
// be set lower than ~ 8kHz. A lower sample frequency would make the
// resampling code overfill its 16k sample ring buffer.
// 
// The end of passband frequency is also limited:
//   pass_freq <= 0.9*sample_freq/2

// E.g. for a 44.1kHz sampling rate the end of passband frequency is limited
// to slightly below 20kHz. This constraint ensures that the FIR table is
// not overfilled.
// ----------------------------------------------------------------------------
void sid_set_sampling_parameters(SIDFP *sid, float clock_freq,
                                 sampling_method method,
                                 float sample_freq)
{
  filter_set_clock_frequency(&sid->filter, clock_freq);
  sid->cycles_per_sample = clock_freq/sample_freq;
  sid->sample_offset = 0;
  sid->sample_prev = 0;
}

// ----------------------------------------------------------------------------
// SID clocking - 1 cycle.
// ----------------------------------------------------------------------------
float sid_clock1(SIDFP *sid)
{
  int i;

  // Clock waveform and envelope generators
  for (i = 0; i < 3; i++) {
    EnvClock(&sid->envelope[i]);
    wave_clock(&sid->wave[i]);
  }

  // Emulate SYNC bit
  wave_synchronize(&sid->wave[0], &sid->wave[1], &sid->wave[2]);
  wave_synchronize(&sid->wave[1], &sid->wave[2], &sid->wave[0]);
  wave_synchronize(&sid->wave[2], &sid->wave[0], &sid->wave[1]);
  
  //env_dac[envelope[0].readENV()] * wave[0].output(wave[2]),
  sid->Vo = (filter_clock(&sid->filter,
    env_dac[EnvRead(&sid->envelope[0])] * 
    wave_output(&sid->wave[0], &sid->wave[2]) + sid->voice_DC,
    env_dac[EnvRead(&sid->envelope[1])] * 
    wave_output(&sid->wave[1], &sid->wave[0]) + sid->voice_DC,
    env_dac[EnvRead(&sid->envelope[2])] * 
    wave_output(&sid->wave[2], &sid->wave[1]) + sid->voice_DC,
    sid->ext_in
  ));

  return sid_output(sid);
}

// ----------------------------------------------------------------------------
// SID clocking - delta_t cycles.
// ----------------------------------------------------------------------------
#if 0
void sid_clock_n(SIDFP *sid, cycle_count delta_t)
{
  int i;
  for (i = 0; i < delta_t; i++) {
    // Clock waveform generators
    wave_clock(&sid->wave[0]);
    wave_clock(&sid->wave[1]);
    wave_clock(&sid->wave[2]);
    // Emulate SYNC bit
    wave_synchronize(&sid->wave[0], &sid->wave[1], &sid->wave[2]);
    wave_synchronize(&sid->wave[1], &sid->wave[2], &sid->wave[0]);
    wave_synchronize(&sid->wave[2], &sid->wave[0], &sid->wave[1]);
  }
}
#endif

// ----------------------------------------------------------------------------
// SID clocking with audio sampling.
//
// The example below shows how to clock the SID a specified amount of cycles
// while producing audio output:
//
// while (delta_t) {
//   bufindex += sid.clock(delta_t, buf + bufindex, buflength - bufindex);
//   write(dsp, buf, bufindex*2);
//   bufindex = 0;
// }
// 
// ----------------------------------------------------------------------------
int sid_clock(SIDFP *sid, cycle_count *delta_t, float* buf, unsigned int n)
{
  int i, s = 0;

  /* XXX I assume n is generally large enough for delta_t here... */
  if (sid->bus_value_ttl != 0) {
    sid->bus_value_ttl -= *delta_t;
    if (sid->bus_value_ttl <= 0) {
        sid->bus_value = 0;
        sid->bus_value_ttl = 0;
    }
  }

  for (;;) {
    float sample_now, next_sample_offset = sid->sample_offset + sid->cycles_per_sample;
    int delta_t_sample = (int)(next_sample_offset);
    if (delta_t_sample > *delta_t) {
      break;
    }
    if ((unsigned int)s >= n) {
      return s;
    }
    for (i = 0; i < delta_t_sample - 1; i++) {
      sid_clock1(sid);
    }
    sid->sample_prev = sid_output(sid);
    sid_clock1(sid);

    *delta_t -= delta_t_sample;
    sid->sample_offset = next_sample_offset - delta_t_sample;

    sample_now = sid_output(sid);
    buf[s++] = (sid->sample_prev + (sid->sample_offset * (sample_now - sid->sample_prev)));
    sid->sample_prev = sample_now;
  }

  for (i = 0; i < *delta_t; i++) {
    sid_clock1(sid);
  }
  sid->sample_prev = sid_output(sid);
  sid->sample_offset -= *delta_t;
  *delta_t = 0;
  return s;
}

