#include <stdint.h>
#include <math.h>
#include "siddefs-fp.h"
#include "wave_c.h"

typedef struct {
    float bias;
    float pulsestrength;
    float topbit;
    float distance;
    float stmix;
} waveformconfig_t;

static float dac[13];
float wftable[11][4096];

void wave_set_nonlinearity(float nl)
{
  float sum, v = 1.0f, w = 1.0f/nl;
  int i;
  dac[11] = sum = 2048.0f*nl;
  for (i = 10; i >= 0; i--) {
    dac[i] = (1U << i) * v;
    sum += dac[i];
    v *= w;
  }
  dac[12] = sum;
}

const float sharpness = 2.0f;
const waveformconfig_t wfconfig[4][5] = {
  { /* kevtris chip G (6581) */
    { 0.5f,       0.0f,      1.0f, 0.0f, 0.0f       },
    { 0.880815f,  0.0f,      0.0f, 0.0f, 0.5999545f }, // error 1795
    { 0.8924618f, 2.014781f, 1.003332f, 0.02992322f, 0.0f }, // error 11610
    { 0.8646501f, 1.712586f, 1.137704f, 0.02845423f, 0.0f }, // error 21307
    { 0.9527834f, 1.794777f, 0.f, 0.09806272f, 0.7752482f }, // error 196
  },
  { /* kevtris chip D (6581r2/r3) */
    { 0.5f,       0.0f,      1.0f,      0.0f,        0.0f        },
    { 0.9347278f, 0.0f,      0.0f,      0.0f,        0.56552734f },
    {0.8931507f, 2.483499f,  0.0f,      0.03339716f, 0.0f},
    {0.8869214f, 2.440879f,  1.680824f, 0.02267573f, 0.0f},
    {0.9266459f, 0.7393153f, 0.0f,      0.0598464f,  0.1851717f},
  },
  { /* kevtris chip G (6581r2/r3) */
    { 0.5f,      0.0f,       1.0f,      0.0f,         0.0f },
    { 0.949378f, 0.0f,       0.0f,      0.0f,         0.6995564f},
    {0.895557f,  2.094213f,  0.0f,      0.028590657f, 0.0f},
    {0.8856111f, 2.299693f,  1.163996f, 0.01210956f,  0.0f},
    {0.9322616f, 0.6173857f, 0.0f,      0.06722359f,  0.2427691f},
  },
  { /* kevtris chip V (8580) */
    {0.5f,     0.0f,    1.0f,     0.0f,     0.0f       },
    {0.5f,     0.0f,    1.0f,     0.0f,     0.9620218f },
    {0.909646f,2.03944f,0.958471f,0.175597f,0.0f    },  // error 18507
    {0.918338f,2.00243f,0.949102f,0.180793f,0.0f    },  // error 16763
    {0.984532f,1.53602f,0.961933f,3.46871f, 0.803955f}, // error 3199
  }
};

/* explode reg12 to a floating point bit array */
static void populate(reg12 v, float o[12])
{
    int i, j = 1;
    for (i = 0; i < 12; i ++) {
        o[i] = v & j ? 1.f : 0.f;
        j <<= 1;
    }
}

/* render output from bitstate */
static float make_sample(float *o) {
  float out = 0;
  for (int i = 0; i < 12; i ++) {
    //out += o[i] * dac[i];
    out += o[i] * (float)i;
  }
  return out;
}

reg8 wave_make_osc_register(float o[12])
{
  reg32 i, bit = 1;
  reg8 out = 0;
  for (i = 4; i < 12; i ++) {
    if (o[i] > 0.5f) {
        out |= bit;
    }
    bit <<= 1;
  }
  return out;
}

reg8 outputN___(reg24 shift_register)
{
  return
    ((shift_register & (1 << 20)) >> 13) |
    ((shift_register & (1 << 18)) >> 12) |
    ((shift_register & (1 << 14)) >> 9) |
    ((shift_register & (1 << 11)) >> 7) |
    ((shift_register & (1 <<  9)) >> 6) |
    ((shift_register & (1 <<  5)) >> 3) |
    ((shift_register & (1 <<  2)) >> 1) |
    ((shift_register & (1 <<  0)));
}

/* waveform values valid are 1 .. 7 */
void calculate_waveform_sample(float o[12], reg24 acc, chip_model model, reg24 my_pw, reg8 wf)
{
  float distancetable[12 * 2 + 1];
  int i, j;

  /* P */
  if (wf == 4) {
    populate((acc >> 12) >= my_pw ? 0xfff : 0x000, o);
    return;
  }

  const waveformconfig_t config = wfconfig[
    model == MOS6581FP ? 2 : 3
  ][
    wf == 3 ? 1 :
    wf == 5 ? 2 :
    wf == 6 ? 3 :
    wf == 7 ? 4 :
              0
  ];

  /* S with strong top bit for 6581 */
  populate(acc >> 12, o);

  /* convert to T */
  if ((wf & 3) == 1) {
    for (i = 11; i > 0; i --) {
      if (acc & 0x800000) {
        o[i] = 1.0f - o[i-1];
      } else {
        o[i] = o[i-1];
      }
    }
    o[0] = 0;
  }

  /* convert to ST */
  if ((wf & 3) == 3) {
    /* bottom bit is grounded via T waveform selector */
    o[0] *= config.stmix;
            /*
             * Enabling the S waveform pulls the XOR circuit selector transistor down
             * (which would normally make the descending ramp of the triangle waveform),
             * so ST does not actually have a sawtooth and triangle waveform combined,
             * but merely combines two sawtooths, one rising double the speed the other.
             *
             * http://www.lemon64.com/forum/viewtopic.php?t=25442&postdays=0&postorder=asc&start=165
             */
    for (i = 1; i < 12; i ++) {
      o[i] = o[i - 1] * (1.f - config.stmix) + o[i] * config.stmix;
    }
    if (model == MOS6581FP)
      o[11] = 0.0f;
  }

  o[11] *= config.topbit;

  /* P* waveform? */
  if (wf > 4) {
    float pulse = (acc >> 12) >= my_pw ? 1.f : -1.f;
    float tmp[12];
    pulse *= config.pulsestrength;

    for (i = 0; i <= 12; i ++) {
        distancetable[12+i] = distancetable[12-i] = 1.f / (1.f + i * i * config.distance);
    }

    for (i = 0; i < 12; i ++) {
        float weight, avg = 0, n = 0;
        for (j = 0; j < 12; j ++) {
            weight = distancetable[i - j + 12];
            avg += o[j] * weight;
            n += weight;
        }
        /* pulse control bit */
        weight = distancetable[i];
        avg += pulse * weight;
        n += weight;
        tmp[i] = (o[i] + avg / n) * 0.5f;
    }

    for (i = 0; i < 12; i ++) {
        o[i] = tmp[i];
    }
  }

  /* use the environment around bias value to set/clear dac bit. The
   * relationship is nonlinear because that seems to sound a bit better.   */
  for (i = 0; i < 12; i ++) {
    float v = (o[i] - config.bias) * sharpness;

    v += 0.5f;
    if (v > 1.0f) {
        v = 1.0f;
    }
    if (v < 0.0f) {
        v = 0.0f;
    }
    o[i] = v;
  }
}

/* generate tables for waveforms 1 .. 7 */
void wave_rebuild_wftable(chip_model model) {
    float wave_zero = (float)(model == MOS6581FP ? -0x380 : -0x800);
    float o[12];
    reg24 acc;
    reg8 wf;

    /* S, T, ST. */
    for (wf = 1; wf < 4; wf ++) {
        for (acc = 0; acc < (1<<24); acc += (1<<12)) {
            calculate_waveform_sample(o, acc, model, 0, wf);
            wftable[wf - 1][acc >> 12] = make_sample(o) + wave_zero;
        }
    }
    /* P */
    for (acc = 0; acc < (1<<12); acc += 1) {
        wftable[3][acc] =    0.0f + wave_zero; /* pulse always low */
        wftable[7][acc] = 4095.0f + wave_zero; /* pulse always high */
    }
    for (wf = 5; wf < 8; wf ++) {
        for (acc = 0; acc < (1<<24); acc += (1<<12)) {
            /* generate pulse-low variants. Also,
             * when waveform < 4, pw doesn't matter. */
            calculate_waveform_sample(o, acc, model, 0x1000, wf); /* pulse always low */
            wftable[wf - 1][acc >> 12] = make_sample(o) + wave_zero;
            /* Add pulse-high variants after pulse-low state variants */
            calculate_waveform_sample(o, acc, model, 0x0000, wf); /* pulse always high */
            wftable[wf + 3][acc >> 12] = make_sample(o) + wave_zero;
        }
    }
}

// ----------------------------------------------------------------------------
// Set chip model.
// ----------------------------------------------------------------------------
void wave_set_chip_model(WaveformGeneratorFP *s, chip_model model)
{
  s->model = model;
  s->wave_zero = (float)(model == MOS6581FP ? -0x380 : -0x800);
}

void wave_reset(WaveformGeneratorFP *s)
{
  s->accumulator_prev = s->accumulator = 0;
  s->previous = 0;
  s->previous_dac = 0;
  s->freq = 0;
  s->pw = 0;
  s->sync = 0;
  s->test = 0;
  s->waveform = 0;
  s->shift_register = 0x7ffffc;
  s->noise_output_cached = 0xfe;
  s->noise_output_cached_dac = dac[12] - 2*dac[4];
}

reg8 wave_readOSC(WaveformGeneratorFP *s, reg24 ring_accumulator, reg24 my_accumulator)
{
  float o[12];
  reg8 out;

  if (s->waveform == 0 || s->waveform > 8) {
    return s->previous;
  }
  if (s->waveform == 8) {
    return (reg8) s->noise_output_cached;
  }

  /* Include effects of the test bit & ring mod */
  my_accumulator ^= (ring_accumulator & s->ring);
  calculate_waveform_sample(o, my_accumulator, s->model, s->test ? 0 : s->pw, s->waveform);
  out = wave_make_osc_register(o);
  return out;
}

reg8 wave_readOSC6581(WaveformGeneratorFP *s,
                      WaveformGeneratorFP *source) {
    return wave_readOSC(s, source->accumulator, s->accumulator);
}

reg8 wave_readOSC8580(WaveformGeneratorFP *s,
                      WaveformGeneratorFP *source) {
    return wave_readOSC(s, source->accumulator_prev, s->accumulator_prev);
}

// ----------------------------------------------------------------------------
// Register functions.
// ----------------------------------------------------------------------------
void wave_writeCONTROL_REG(WaveformGeneratorFP *s, WaveformGeneratorFP* source, reg8 control)
{
  reg8 test_next = 0;
  /* when selecting the 0 waveform, the previous output is held for
   * a time in the DAC MOSFET gates. We keep on holding forever, though... */
  if ((control & 0x10) == 0 && s->waveform >= 1 && s->waveform <= 8) {
    /* NB: "6581" version is the undelayed variant. We want that. */
    s->previous = wave_readOSC6581(s, source);
    s->previous_dac = wave_output(s, source);
  }

  s->waveform = (control >> 4) & 0x0f;
  s->ring = (control & 0x04) != 0 && (s->waveform & 0x3) == 1 ? 0x800000 : 0;
  s->sync = ((control & 0x02) >> 1);
  test_next = ((control & 0x08) >> 2);

  // Test bit rising? Invert bit 19 and write it to bit 1.
  if (test_next && !s->test) {
    s->accumulator = 0;
    s->noise_overwrite_delay = 200000; /* 200 ms, probably too generous? */
  } else {
    // Test bit falling? clock noise once,
    // otherwise just emulate noise's combined waveforms.
    if (! test_next) {
      wave_clock_noise(s, s->test);
    }
  }
 
  s->test = test_next;
}

void wave_clock_noise(WaveformGeneratorFP *s, const uint8_t clock)
{
  int i;

  if (clock) {
    reg24 bit0 = (((s->shift_register >> 22) | (s->test ? 1 : 0)) ^ (s->shift_register >> 17)) & 0x1;
    s->shift_register <<= 1;
    s->shift_register |= bit0;
  }

  // clear output bits of shift register if noise and other waveforms
  // are selected simultaneously
  if (s->waveform > 8) {
    s->shift_register &= 0x7fffff^(1<<20)^(1<<18)^(1<<14)^(1<<11)^(1<<9)^(1<<5)^(1<<2)^(1<<0);
  }

  s->noise_output_cached = outputN___(s->shift_register);
  s->noise_output_cached_dac = s->wave_zero;
  for (i = 0; i < 8; i ++) {
    if (s->noise_output_cached & (1 << i)) {
      s->noise_output_cached_dac += dac[i+4];
    }
  }
}

// ----------------------------------------------------------------------------
// Select one of 16 possible combinations of waveforms.
// ----------------------------------------------------------------------------
float wave_output(WaveformGeneratorFP* s, WaveformGeneratorFP* sync_source)
{
  reg24 output_pw;
  reg8 waveform = s->waveform;
  int variant;

  if (waveform == 0 || waveform > 8) {
    return s->previous_dac;
  }
  if (waveform == 8) {
    return s->noise_output_cached_dac;
  }
  /* waveforms 1 .. 7 left */

  /* pulse on/off generates 4 more variants after the main pulse types */
  /* Pulse not used? Always use the -1 variant */
  //if ((waveform & 4) == 0) {
    //output_pw = 1 << 24;
  //} else {
    /* Test bit -> always on, otherwise compare pw to accumulator */
    output_pw = s->pw << 12;
  //}
  //variant = accumulator >= output_pw ? 3 : -1;
  variant = ((waveform & 4) && (s->accumulator >= output_pw)) ? 4 : 0;

  /* triangle waveform XOR circuit. Since the table already makes a triangle
   * wave internally, we only need to account for the sync source here.
   * Flipping the top bit suffices to reproduce the original SID ringmod */
   reg24 phase = s->accumulator ^ (sync_source->accumulator & s->ring);
  return wftable[waveform - 1 + variant][phase >> 12];
}

