//  ---------------------------------------------------------------------------
//  This file is part of reSID, a MOS6581 SID emulator engine.
//  Copyright (C) 2004  Dag Lem <resid@nimrod.no>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//  ---------------------------------------------------------------------------

#ifndef VICE__WAVE_H__
#define VICE__WAVE_H__

#include "siddefs-fp.h"

// ----------------------------------------------------------------------------
// A 24 bit accumulator is the basis for waveform generation. FREQ is added to
// the lower 16 bits of the accumulator each cycle.
// The accumulator is set to zero when TEST is set, and starts counting
// when TEST is cleared.
// The noise waveform is taken from intermediate bits of a 23 bit shift
// register. This register is clocked by bit 19 of the accumulator.
// ----------------------------------------------------------------------------
typedef struct WaveformGeneratorFP
{
  chip_model model;

  reg24 accumulator, accumulator_prev;
  reg24 shift_register;
  reg12 noise_output_cached;
  reg8 previous;
  reg24 noise_overwrite_delay;

  // Fout  = (Fn*Fclk/16777216)Hz
  reg16 freq;
  // PWout = (PWn/40.95)%
  reg12 pw;

  // The control register right-shifted 4 bits; used for output function
  // table lookup.
  reg8 waveform;

  // The remaining control register bits.
  reg8 test, sync;
  // The gate bit is handled by the EnvelopeGenerator.

  /* Ring mod xor gate. */
  reg24 ring;

  // zero level offset of waveform (< 0)
  float wave_zero;
  float previous_dac, noise_output_cached_dac;
} WaveformGeneratorFP;

#ifdef __cplusplus
extern "C" {
#endif

void wave_set_nonlinearity(float nl);
void wave_rebuild_wftable(chip_model model);
void calculate_waveform_sample(float o[12], reg24 acc, chip_model model, reg24 my_pw, reg8 wf);
void wave_set_chip_model(WaveformGeneratorFP *s, chip_model model);
void wave_reset(WaveformGeneratorFP *s);
reg8 wave_readOSC(WaveformGeneratorFP *s, reg24 ring_accumulator, reg24 my_accumulator);
reg8 wave_readOSC6581(WaveformGeneratorFP *s, WaveformGeneratorFP *source);
reg8 wave_readOSC8580(WaveformGeneratorFP *s, WaveformGeneratorFP *source);
void wave_writeCONTROL_REG(WaveformGeneratorFP *s, WaveformGeneratorFP* source, reg8 control);
void wave_clock_noise(WaveformGeneratorFP *s, const uint8_t clock);
float wave_output(WaveformGeneratorFP* s, WaveformGeneratorFP* sync_source);

// ----------------------------------------------------------------------------
// Register functions.
// ----------------------------------------------------------------------------
inline void wave_writeFREQ_LO(WaveformGeneratorFP *s, reg8 freq_lo)
{
  s->freq = (s->freq & 0xff00) | (freq_lo & 0x00ff);
}

inline void wave_writeFREQ_HI(WaveformGeneratorFP *s, reg8 freq_hi)
{
  s->freq = ((freq_hi << 8) & 0xff00) | (s->freq & 0x00ff);
}

inline void wave_writePW_LO(WaveformGeneratorFP *s, reg8 pw_lo)
{
  s->pw = ((s->pw & 0xf00) | (pw_lo & 0x0ff));
}

inline void wave_writePW_HI(WaveformGeneratorFP *s, reg8 pw_hi)
{
  s->pw = (((pw_hi << 8) & 0xf00) | (s->pw & 0x0ff));
}

// ----------------------------------------------------------------------------
// SID clocking - 1 cycle.
// ----------------------------------------------------------------------------
static inline void wave_clock(WaveformGeneratorFP *s)
{
  /* no digital operation if test bit is set. Only emulate analog fade. */
  if (s->test) {
    if (s->noise_overwrite_delay != 0) {
	  if (-- s->noise_overwrite_delay == 0) {
	    s->shift_register |= 0x7ffffc;
	    wave_clock_noise(s, 0);
      }
    }
    return;
  }

  s->accumulator_prev = s->accumulator;

  // Calculate new accumulator value;
  s->accumulator += s->freq;
  s->accumulator &= 0xffffff;

  // Shift noise register once for each time accumulator bit 19 is set high.
  //if (!(accumulator_prev & 0x080000) && (accumulator & 0x080000)) {
  if ((~s->accumulator_prev & s->accumulator) & 0x080000U) {
    wave_clock_noise(s, 1);
  }
}

// ----------------------------------------------------------------------------
// SID clocking minus noise waveform - 1 cycle.
// ----------------------------------------------------------------------------
static inline void wave_clock_fast(WaveformGeneratorFP *s)
{
  /* no digital operation if test bit is set. */
  if (!s->test) {
    // Calculate new accumulator value;
    s->accumulator_prev = s->accumulator;
    s->accumulator += s->freq;
    s->accumulator &= 0xffffff;
  }
}

// ----------------------------------------------------------------------------
// Synchronize oscillators.
// This must be done after all the oscillators have been clock()'ed since the
// oscillators operate in parallel.
// Note that the oscillators must be clocked exactly on the cycle when the
// MSB is set high for hard sync to operate correctly. See SIDFP::clock().
// ----------------------------------------------------------------------------
static inline void wave_synchronize(WaveformGeneratorFP *s, WaveformGeneratorFP* sync_dest, WaveformGeneratorFP* sync_source)
{
  // A special case occurs when a sync source is synced itself on the same
  // cycle as when its MSB is set high. In this case the destination will
  // not be synced. This has been verified by sampling OSC3.
  if ( (sync_dest->sync && ((~s->accumulator_prev)           & s->accumulator           & 0x800000) != 0) &&
      !(s->sync         && ((~sync_source->accumulator_prev) & sync_source->accumulator & 0x800000) != 0)) {
    sync_dest->accumulator = 0;
  }
}

#ifdef __cplusplus
};
#endif

#endif // not VICE__WAVE_H__
